#include <ArduinoBLE.h>
#include "Arduino_NineAxesMotion.h"
#include <Wire.h>
#include <Arduino_LSM6DS3.h>

/*  This program uses two libraries: ArduinoBLE and Arduino_NineAxesMotion.
    These libraries are important for understanding this program.
    Examples located in the github repos are  great for learnings.
    https://www.arduino.cc/en/Reference/ArduinoBLE,
    https://github.com/arduino-libraries/ArduinoBLE
    https://github.com/arduino-libraries/Arduino_NineAxesMotion
    https://www.arduino.cc/en/Reference/ArduinoLSM6DS3
*/

// Holds the values read from the strain gauge sensor
union strain_gauge {
  struct __attribute__((packed)) {
    unsigned long timer;
    short gauge1;
    short gauge2;
    short gauge3;
    short gauge4;
    short gauge5;
    short gauge8;
  };
  byte bytes[16];
};

// Holds the values read from the IMU sensor
union acc_gyro {
  struct __attribute__((packed)) {
    unsigned long timer;
    short acc_x;
    short acc_z;
    short gyro_y;
  };
  byte bytes[10];
};

union strain_gauge strain_gauge_data; // Stores strain gauge values
union acc_gyro acc_gyro_data;         // Stores IMU values
union acc_gyro grav_gyro_data;        // Stores IMU values

NineAxesMotion IMUSensor; // The IMUSensor object for reading and updating the 9-axis motion shield mounted on the Arduino.
BLEService structural_health_data_Service("19b10000-e8f2-537e-4f6c-d104768a1214"); // Creating the bluetooth service

// Creating the bluetooth characteristics
BLECharacteristic strain_gauge_data_characteristic("19b10001-e8f2-537e-4f6c-d104768a1214", BLERead | BLENotify, sizeof(strain_gauge_data));
BLECharacteristic acc_gyro_data_characteristic("19b10002-e8f2-537e-4f6c-d104768a1214", BLERead | BLENotify, sizeof(acc_gyro_data));
BLECharacteristic grav_gyro_data_characteristic("19b10003-e8f2-537e-4f6c-d104768a1214", BLERead | BLENotify, sizeof(grav_gyro_data));

//Pins used for the LED
int redLEDPin = 5;
int greenLEDPin = 4;
int blueLEDPin = 3;

// Timer values for the LED blink functionality
long blinkTimer = 0;
long previousMillis = 0;

// Sample hold time in milliseconds. Value of 10 gives 1000/10 = 100 Hz sample frequency.
unsigned long sampleAndHoldTime = 10; // 

void setup() {
  //Serial.begin(9600);  //Un-comment if you wish to read value from the serial monitor.

  // Setting output pin modes for the LEDs.
  pinMode(redLEDPin, OUTPUT);
  pinMode(greenLEDPin, OUTPUT);
  pinMode(blueLEDPin, OUTPUT);

  // Setting input pin modes.
  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);
  pinMode(A3, INPUT);
  pinMode(A4, INPUT);
  pinMode(A5, INPUT);

  Wire.begin(); //Initialize I2C communication to the let the library communicate with the sensor.
  //Sensor Initialization
  IMUSensor.initSensor(); //The I2C Address can be changed here inside this function in the library
  IMUSensor.setOperationMode(OPERATION_MODE_NDOF);  //Can be configured to other operation modes as desired
  IMUSensor.setUpdateMode(MANUAL);  //The default is AUTO. Changing to manual requires calling the relevant update functions prior to calling the read functions
  //Setting to MANUAL requires lesser reads to the sensor
  IMUSensor.updateAccelConfig();

  if (!BLE.begin()) {
    //Serial.println("BLE initialization failed!"); // Un-comment if using serial.
    while (1);
  }

  if (!IMU.begin()) {
    //Serial.println("BLE initialization failed!"); // Un-comment if using serial.
    while (1);
  }
  
  //Serial.println("BLE Central - Peripheral Explorer");  Un-comment if using serial.
  // Bluetooth Setup
  BLE.setLocalName("SHM - UNO_WIFI");
  BLE.setAdvertisedService(structural_health_data_Service);
  structural_health_data_Service.addCharacteristic(acc_gyro_data_characteristic);
  structural_health_data_Service.addCharacteristic(strain_gauge_data_characteristic);
  structural_health_data_Service.addCharacteristic(grav_gyro_data_characteristic);
  BLE.addService(structural_health_data_Service);

  BLE.advertise();
  //Serial.println("Bluetooth device active, waiting for connections...");  // Un-comment if using serial.

  blinkTimer = millis();
  digitalWrite(redLEDPin, HIGH);
}

long startTime = -1;
/* The Loop method runs after the setup is complete.
   Method listens for BLE peripherals, and updates the values only if one is connected.
   Values are updated every 10 ms.
*/
void loop() {
  BLEDevice central = BLE.central();

  if (central) {
    //Serial.print("Connected to central: ");
    //Serial.println(central.address());

    digitalWrite(blueLEDPin, HIGH);  // Turn on blue LED

    while (central.connected()) {

        unsigned long currentMillis = millis();

        // Check for any characteristic subscribers
        if(strain_gauge_data_characteristic.subscribed() or acc_gyro_data_characteristic.subscribed() or grav_gyro_data_characteristic.subscribed()){
            if(startTime == -1){
              startTime = millis();
            }

             // Update sensor values when difference is larger than the sample hold time
            if (currentMillis - previousMillis >= sampleAndHoldTime) {
                previousMillis = currentMillis;
                ledBlink(greenLEDPin, 1); // Blink green LED

                if(strain_gauge_data_characteristic.subscribed()){
                  updateStrainGaugeValues(currentMillis - startTime);
                }
                if(acc_gyro_data_characteristic.subscribed()){
                  updateAccGyroValuesBNO(currentMillis - startTime);
                  //updateAccGyroValuesLSM(currentMillis - startTime);
                }

                if(grav_gyro_data_characteristic.subscribed()){
                  updateGravGyroValues(currentMillis - startTime);
                }
            }
        }
    }
    
    startTime = -1; // Reset the timer
    digitalWrite(greenLEDPin, LOW); // Turn off green LED
    digitalWrite(blueLEDPin, LOW); // Turn off blue LED
  }
}

/* Turns LED on and off every 1000 ms.
   ledPin: LED output pin, timer in milliseconds.
   timer : timer in milliseconds.
   scalar: scalar for adjusting frequency multiplicatively.
*/
void ledBlink(int ledPin, float scalar) {
  if (millis() - blinkTimer > 2000 * scalar) {
    digitalWrite(ledPin, HIGH);
    blinkTimer = millis();
  }
  else if (millis() - blinkTimer > 1000 * scalar) {
    digitalWrite(ledPin, LOW);
  }
}

// Reads the strain gauge sensors, stores them, and writes them to the Bluetooth characteristic.
// The bluetooth characteristic publishes on write.
void updateStrainGaugeValues(float timeStamp) {
  strain_gauge_data.timer = timeStamp;
  strain_gauge_data.gauge1 = analogRead(A0);
  strain_gauge_data.gauge2 = analogRead(A1);
  strain_gauge_data.gauge3 = analogRead(A2);
  strain_gauge_data.gauge4 = analogRead(A3);
  strain_gauge_data.gauge5 = analogRead(A4);
  strain_gauge_data.gauge8 = analogRead(A5);
  strain_gauge_data_characteristic.writeValue(strain_gauge_data.bytes, 16); //sizeof(strain_gauge_data)
}


// Reads the acceleration and gyroscope of the BNO055 IMU, stores them, and writes them to the Bluetooth characteristic.
// The bluetooth characteristic publishes on write.
void updateGravGyroValues(float timeStamp) {
    IMUSensor.updateGyro();
    IMUSensor.updateGravAccel();          //Update the Gravity Acceleration data
    IMUSensor.updateCalibStatus();
    grav_gyro_data.timer = timeStamp;
    grav_gyro_data.gyro_y = IMUSensor.readGyroX() * 100;
    grav_gyro_data.acc_x = IMUSensor.readGravAcceleration(X_AXIS) * 100;
    grav_gyro_data.acc_z  = IMUSensor.readGravAcceleration(Z_AXIS) * 100;
    grav_gyro_data_characteristic.writeValue(grav_gyro_data.bytes, 10); //sizeof(grav_gyro_data)
}


float ax,ay,az;
float gyx,gyy,gyz;
// Reads the acceleration and gyroscope of the LSM IMU, stores them, and writes them to the Bluetooth characteristic.
// The bluetooth characteristic publishes on write.
void updateAccGyroValuesLSM(float timeStamp) {
    IMU.readGyroscope(gyx, gyy, gyz);
    IMU.readAcceleration(ax, ay, az);
    acc_gyro_data.timer = timeStamp;
    acc_gyro_data.acc_x = short(ax*100);
    acc_gyro_data.acc_z = short(az*100);
    acc_gyro_data.gyro_y = short(gyy*100);
    acc_gyro_data_characteristic.writeValue(acc_gyro_data.bytes, 10); //sizeof(grav_gyro_data)
}

// Updates the IMUs, reads the acceleration and gyroscope of the BNO055 IMU, stores them, and writes them to the Bluetooth characteristic.
// The bluetooth characteristic publishes on write.
void updateAccGyroValuesBNO(float timeStamp) {
    IMUSensor.updateGyro();
    IMUSensor.updateAccel();          //Update the Gravity Acceleration data
    IMUSensor.updateCalibStatus();
    acc_gyro_data.timer = timeStamp;
    acc_gyro_data.gyro_y = IMUSensor.readGyroX() * 100;
    acc_gyro_data.acc_x = IMUSensor.readAccelerometer(X_AXIS) * 100;
    acc_gyro_data.acc_z  = IMUSensor.readAccelerometer(Z_AXIS) * 100;
    acc_gyro_data_characteristic.writeValue(acc_gyro_data.bytes, 10); //sizeof(grav_gyro_data)
}
